# -*- coding: utf-8 -*-

import os
import sys
import re
import shlex
import subprocess

ENV_PATTERN = re.compile('^(?P<name>[^\s]+)\s*=\s*(?P<value>.+)$')


def previous_alias(alias):
    return '{}-pre'.format(alias)


def parse_env_args(env):
    if not isinstance(env, list):
        return None

    ret = {}
    for e in env:
        matches = ENV_PATTERN.match(e)

        if matches is None:
            raise Exception(
                'The format of "env" option must be "NAME=VALUE": {}'.format(e))

        name = matches.group('name')
        value = matches.group('value')
        k, v = shlex.split('{} {}'.format(name, value))
        ret[k] = v

    return ret


def run_commands(commands, working_dir=os.getcwd()):
    cwd = os.getcwd()
    os.chdir(working_dir)

    for c in commands:
        try:
            subprocess.check_output(
                c, stderr=subprocess.STDOUT, shell=True)
        except subprocess.CalledProcessError as e:
            os.chdir(cwd)
            raise Exception(e.output)

    os.chdir(cwd)


def confirm_overwrite(path):
    ret = True
    if os.path.exists(path):
        print('Overwrite {}? [y/n]: '.format(path))
        y_n = sys.stdin.readline()
        if not y_n.startswith('y'):
            ret = False
    return ret



# -*- coding: utf-8 -*-

import time
import signal

from time import mktime
from parsedatetime import Calendar
from lamvery.actions.base import BaseAction


class LogsAction(BaseAction):

    def __init__(self, args):
        super(LogsAction, self).__init__(args)
        self._follow = args.follow
        self._filter = args.filter
        self._interval = args.interval
        self._exit = False
        self._start = args.start

    def action(self):
        self._logger.info('Start viewing the log events...')

        def _exit(signum, frame):
            self._logger.info('Exit by code {} ...'.format(signum))
            self._exit = True

        signal.signal(signal.SIGTERM, _exit)
        signal.signal(signal.SIGINT, _exit)

        start = time.time()
        if self._start is not None:
            time_struct, _ = Calendar().parse(self._start)
            start = mktime(time_struct)

        start = int(start * 1000)

        client = self.get_logs_client()
        function = self._config.get_function_name()
        event_ids = {}

        while self._exit is False:
            events = client.get_log_events(function, start, self._filter)

            for e in events:
                if e['eventId'] not in event_ids:
                    event_ids[e['eventId']] = None
                    print(e['message'])

                    if e['timestamp'] > start:
                        start = e['timestamp']
                        event_ids = {}

            if not self._follow:
                break

            time.sleep(self._interval)



# -*- coding: utf-8 -*-

import botocore

from lamvery.clients.base import BaseClient


class EventsClient(BaseClient):

    def __init__(self, *args, **kwargs):
        super(EventsClient, self).__init__(*args, **kwargs)
        self._events = self._session.client('events')

    def get_rules_by_target(self, arn):
        names = self._get_rule_names_by_tagert(arn=arn)

        rules = []
        for name in names:
            rules.append(self._events.describe_rule(Name=name))

        return rules

    def _get_rule_names_by_tagert(self, arn, next_token=None):
        try:
            if next_token:
                names = self._events.list_rule_names_by_target(
                    TargetArn=arn, NextToken=next_token, Limit=100)
            else:
                names = self._events.list_rule_names_by_target(
                    TargetArn=arn, Limit=100)
        except botocore.exceptions.ClientError:
            return []

        ret = names['RuleNames']
        if 'NextToken' in names:
            ret.extend(
                self._get_rule_names_by_tagert(
                    arn=arn, next_token=names['NextToken']))
        return ret

    def put_rule(self, rule):
        if self._dry_run:
            return {}
        else:
            kwargs = {}
            kwargs['Name'] = rule['name']
            kwargs['State'] = rule.get('state', 'ENABLED')

            if rule.get('description') is not None:
                kwargs['Description'] = rule.get('description')
            if rule.get('pattern') is not None:
                kwargs['EventPattern'] = rule.get('pattern')
            if rule.get('schedule') is not None:
                kwargs['ScheduleExpression'] = rule.get('schedule')

            return self._events.put_rule(**kwargs)

    def put_targets(self, rule, targets, arn):
        if not self._dry_run:
            t_args = []

            for t in targets:
                arg = {'Id': t['id'], 'Arn': arn}
                if t.get('input') is not None:
                    arg['Input'] = t['input']
                if t.get('input_path') is not None:
                    arg['InputPath'] = t['input_path']
                t_args.append(arg)

            self._events.put_targets(Rule=rule, Targets=t_args)

    def get_targets_by_rule(self, rule, next_token=None):
        try:
            if next_token:
                rules = self._events.list_targets_by_rule(
                    Rule=rule, NextToken=next_token, Limit=100)
            else:
                rules = self._events.list_targets_by_rule(
                    Rule=rule, Limit=100)
        except botocore.exceptions.ClientError:
            return []

        ret = rules['Targets']
        if 'NextToken' in rules:
            ret.extend(
                self.get_targets_by_rule(
                    rule=rule, next_token=rules['NextToken']))
        return ret

    def remove_targets(self, rule, targets):
        if not self._dry_run:
            self._events.remove_targets(Rule=rule, Ids=targets)

    def delete_rule(self, name):
        if not self._dry_run:
            self._events.delete_rule(Name=name)



# -*- coding: utf-8 -*-

import botocore
import hashlib

from lamvery.clients.base import BaseClient
from lamvery.utils import previous_alias


class LambdaClient(BaseClient):

    def __init__(self, *args, **kwargs):
        super(LambdaClient, self).__init__(*args, **kwargs)
        self._lambda = self._session.client('lambda')

    def get_function_conf(self, name, alias=None):
        try:
            kwargs = {}
            kwargs['FunctionName'] = name

            if alias is not None:
                kwargs['Qualifier'] = alias

            res = self._lambda.get_function(**kwargs)
            return res['Configuration']
        except botocore.exceptions.ClientError:
            return {}

    def create_function(self, zipfile, conf, publish):
        kwargs = {}
        kwargs['FunctionName'] = conf['name']
        kwargs['Runtime'] = conf['runtime']
        kwargs['Role'] = conf['role']
        kwargs['Handler'] = conf['handler']
        kwargs['Code'] = {'ZipFile': zipfile.read()}
        kwargs['Publish'] = publish

        description = conf.get('description')
        if description is not None:
            kwargs['Description'] = description

        timeout = conf.get('timeout')
        if timeout is not None:
            kwargs['Timeout'] = timeout

        memory_size = conf.get('memory_size')
        if memory_size is not None:
            kwargs['MemorySize'] = memory_size

        vpc_config = conf.get('vpc_config')
        if vpc_config is not None:
            kwargs['VpcConfig'] = self._build_vpc_config(vpc_config)

        environment_variables = conf.get('environment_variables')
        if environment_variables is not None:
            kwargs['Environment'] = {'Variables': {}}
            kwargs['Environment']['Variables'] = conf['environment_variables']

        if not self._dry_run:
            self._lambda.create_function(**kwargs)

    def _build_vpc_config(self, vpc_config):
        return {
            'SubnetIds': vpc_config['subnets'],
            'SecurityGroupIds': vpc_config['security_groups']}

    def update_function_code(self, zipfile, conf, publish):
        if not self._dry_run:
            ret = self._lambda.update_function_code(
                FunctionName=conf['name'],
                ZipFile=zipfile.read(),
                Publish=publish)
            return ret['Version']
        return None

    def update_function_conf(self, conf):
        kwargs = {}
        kwargs['FunctionName'] = conf['name']
        kwargs['Role'] = conf['role']
        kwargs['Handler'] = conf['handler']

        description = conf.get('description')
        if description is not None:
            kwargs['Description'] = description

        timeout = conf.get('timeout')
        if timeout is not None:
            kwargs['Timeout'] = timeout

        memory_size = conf.get('memory_size')
        if memory_size is not None:
            kwargs['MemorySize'] = memory_size

        vpc_config = conf.get('vpc_config')
        if vpc_config is not None:
            kwargs['VpcConfig'] = self._build_vpc_config(vpc_config)

        environment_variables = conf.get('environment_variables')
        if environment_variables is not None:
            kwargs['Environment'] = {'Variables': {}}
            kwargs['Environment']['Variables'] = conf['environment_variables']

        if not self._dry_run:
            self._lambda.update_function_configuration(**kwargs)

    def get_alias(self, function, alias):
        try:
            return self._lambda.get_alias(
                FunctionName=function,
                Name=alias)
        except botocore.exceptions.ClientError:
            return {}

    def create_alias(self, function, alias, version):
        if not self._dry_run:
            self._lambda.create_alias(
                FunctionName=function,
                Name=alias,
                FunctionVersion=version)

    def update_alias(self, function, alias, version):
        if not self._dry_run:
            self._lambda.update_alias(
                FunctionName=function,
                Name=alias,
                FunctionVersion=version)

    def calculate_capacity(self, next_marker=None):
        if next_marker:
            r = self._lambda.list_functions(MaxItems=500, Marker=next_marker)
        else:
            r = self._lambda.list_functions(MaxItems=500)

        size = sum(
            self._calculate_versions_capacity(
                f['FunctionName']) for f in r['Functions'])

        if 'NextMarker' in r:
            return size + self.calculate_capacity(next_marker=r['NextMarker'])
        else:
            return size

    def _calculate_versions_capacity(self, function_name, next_marker=None):
        if next_marker:
            r = self._lambda.list_versions_by_function(
                FunctionName=function_name, MaxItems=500, Marker=next_marker)
        else:
            r = self._lambda.list_versions_by_function(
                FunctionName=function_name, MaxItems=500)

        size = sum(f['CodeSize'] for f in r['Versions'])

        if 'NextMarker' in r:
            return size + self._calculate_versions_capacity(
                function_name=function_name, next_marker=r['NextMarker'])
        else:
            return size

    def add_permission(self, function, alias, name, arn, principal='events.amazonaws.com'):
        kwargs = {}
        kwargs['FunctionName'] = function
        kwargs['Action'] = 'lambda:InvokeFunction'
        kwargs['Principal'] = principal
        kwargs['SourceArn'] = arn
        kwargs['StatementId'] = self._generate_statement_id(function, name, alias)

        if alias is not None:
            kwargs['Qualifier'] = alias

        if not self._dry_run:
            try:
                self._lambda.add_permission(**kwargs)
            except botocore.exceptions.ClientError:
                pass

    def remove_permission(self, function, alias, rule):
        kwargs = {}
        kwargs['FunctionName'] = function
        kwargs['StatementId'] = self._generate_statement_id(function, rule, alias)

        if alias is not None:
            kwargs['Qualifier'] = alias

        if not self._dry_run:
            self._lambda.remove_permission(**kwargs)

    def _generate_statement_id(self, function, rule, alias):
        return hashlib.sha256(
            'lamvery-{}-{}-{}'.format(function, rule, alias)).hexdigest()

    def invoke(self, name, qualifier=None, payload=None):
        kwargs = {}
        kwargs['FunctionName'] = name
        kwargs['InvocationType'] = 'RequestResponse'
        kwargs['LogType'] = 'Tail'

        if payload is not None:
            kwargs['Payload'] = payload
        if qualifier is not None:
            kwargs['Qualifier'] = qualifier

        return self._lambda.invoke(**kwargs)

    def get_previous_version(self, function, alias):
        ver = self.get_alias(function, previous_alias(alias))
        return ver.get('FunctionVersion')



"""
lamvery
----
Function based deploy and management tool for AWS Lambda.
"""

__version__ = '0.18.0'

import importlib

secret = importlib.import_module('lamvery.secret')
env = importlib.import_module('lamvery.env')



# -*- coding: utf-8 -*-

from lamvery.clients.base import BaseClient


class LogsClient(BaseClient):

    def __init__(self, *args, **kwargs):
        super(LogsClient, self).__init__(*args, **kwargs)
        self._logs = self._session.client('logs')

    def get_log_events(self, function, start, pattern, next_token=None):
        kwargs = {}
        kwargs['logGroupName'] = '/aws/lambda/{}'.format(function)
        kwargs['startTime'] = start
        kwargs['limit'] = 100

        if pattern is not None:
            kwargs['filterPattern'] = pattern
        if next_token is not None:
            kwargs['nextToken'] = next_token

        log_events = self._logs.filter_log_events(**kwargs)

        ret = log_events['events']
        if 'NextToken' in log_events:
            ret.extend(
                self.get_log_events(
                    function, start, pattern, log_events['NextToken']))
        return ret



# -*- coding: utf-8 -*-

import boto3

from abc import ABCMeta


class BaseClient:

    __metaclass__ = ABCMeta

    def __init__(self, region=None, profile=None, dry_run=False):
        self._session = boto3.session.Session(
            profile_name=profile, region_name=region)
        self._dry_run = dry_run
        self._sts = self._session.client('sts')

    def get_account_id(self):
        return self._sts.get_caller_identity().get('Account')



# -*- coding: utf-8 -*-

import botocore
import json

from lamvery.clients.base import BaseClient


class ApiGatewayClient(BaseClient):

    def __init__(self, *args, **kwargs):
        super(ApiGatewayClient, self).__init__(*args, **kwargs)
        self._api = self._session.client('apigateway')

    def get_rest_api(self, api_id):
        if api_id is None:
            return None

        try:
            return self._api.get_rest_api(restApiId=api_id)
        except botocore.exceptions.ClientError:
            return None

    def get_export(self, api_id, stage):
        if api_id is None:
            return {}

        try:
            ret = self._api.get_export(
                restApiId=api_id,
                stageName=stage,
                exportType='swagger',
                parameters={'extensions': 'integrations'})
            return json.loads(ret['body'].read())

        except botocore.exceptions.ClientError:
            return {}

    def import_rest_api(self, api_conf):
        return self._api.import_rest_api(body=json.dumps(api_conf))

    def put_rest_api(self, api_id, api_conf):
        return self._api.put_rest_api(
            restApiId=api_id,
            mode='overwrite',
            body=json.dumps(api_conf))

    def delete_rest_api(self, api_id):
        self._api.delete_rest_api(api_id)

    def create_deployment(self, api_id, stage):
        return self._api.create_deployment(
            restApiId=api_id,
            stageName=stage)
